
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_active_db.h"

#include "initng_envparser.h"
s_entry ENV_FILE = { "env_file", STRINGS, 8, NULL,
    "Parse this file for environmental variables."
};
s_entry ENV_FILE_REQUIRED = { "env_file_required", STRINGS, 17, NULL,
    "Same as env_file, but service will fail if the file doesn't exist."
};

static int parse_file(const char *file, service_h * s);

/* Parse all the env_file's when service finished loading */
static int env_parser(service_h * s)
{
    const char *file = NULL;

    assert(s);

    /* make sure the env_file is set */
    if (!service_db_is(&ENV_FILE, s))
        return (TRUE);

    /* Parse all ENV_FILE's */
    while ((file = service_db_get_next_string(&ENV_FILE, s, file)))
    {
        parse_file(file, s);
    }

    /* Parse all, and stop if files dont exits, or not parseable */
    while ((file = service_db_get_next_string(&ENV_FILE_REQUIRED, s, file)))
    {
        if (parse_file(file, s) == FALSE)
            return (FALSE);
    }

    return (TRUE);
}

static int parse_file(const char *file, service_h * s)
{
    int i = 0;
    int j = 0;

    char *file_content = NULL;
    char *point = NULL;

    /* open that file */
    if (!open_read_close(file, &file_content))
        return (FALSE);


    /* Start Parsing */
    point = file_content;
    while (point[0])
    {
        /* skip initzial spaces */
        JUMP_SPACES(point);
        if (!point[0])
            break;

        /* skip rows, starting with '#' */
        if (point[0] == '#')
        {
            JUMP_TO_NEXT_ROW(point);
            continue;
        }

        /* make sure the line contains a '=', if not, its not an ENV line */
        i = 0;
        while (point[i] && point[i] != '=' && point[i] != '\n')
            i++;
        if (point[i] != '=')
        {
            JUMP_TO_NEXT_ROW(point);
            continue;
        }

        /* count rowlen */
        j = i;                              /* minimum */
        while (point[j] && point[j] != '\n')
            j++;

        /*
         * a line like OPT="fist second third" should not be added with
         * the fisrst and last '"'
         * change the line from:
         * OPT="first second third\0"
         * to:
         * OPT=first second third\0
         */
        if (point[i + 1] == '"' && point[j - 1] == '"')
        {
            point[i + 1] = point[i + 2];
            int k = 0;

            while (point[i + 2 + k] && point[i + 2 + k] != '"'
                   && point[i + 2 + k] != '\n')
            {
                point[i + 1 + k] = point[i + 2 + k];
                k++;
            }
            /* degrese one for every of the two '"' */
            j--;
            j--;
            point[j] = '\0';
        }
        /* duplicate, and add this line */
        service_db_set_another_string(&ENV, s, i_strndup(point, j));

        /* go to next row and parse */
        JUMP_TO_NEXT_ROW(point);
    }
    free(file_content);
    return (TRUE);
}

int module_init(const char *version)
{
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }
    D_("module_init();\n");

    initng_sdt_add(&ENV_FILE);
    initng_sdt_add(&ENV_FILE_REQUIRED);
    return (initng_add_hook(ADDITIONAL_PARSE, 80, &env_parser));
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&ENV_FILE);
    initng_sdt_del(&ENV_FILE_REQUIRED);
    initng_del_hook(ADDITIONAL_PARSE, &env_parser);
}
