
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <dirent.h>   /*#include <dirent.h> */  /* opendir() closedir() */
#include <fnmatch.h>

#include "../../src/initng_global.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_error.h"
#include "../../src/initng_handler.h"
#include "../../initng-paths.h"

#include "initng_i_parser.h"

#define ENDED(x) (!(x)[0] || (x)[0]=='\n')
char *g_filename = NULL;
char *g_pointer = NULL;

#define print_line(a,m) err_print_line2t(a, m, MSG, __FILE__, (const char*)__PRETTY_FUNCTION__, __LINE__)
#define err_print_line(a,m) err_print_line2t(a, m, MSG_FAIL, __FILE__, (const char*)__PRETTY_FUNCTION__, __LINE__)
static void err_print_line2t(char *point, const char *message, e_mt err,
                             const char *file, const char *func,
                             int codeline);
static service_h *parse_file(char *filename, const char *watch_for);
static int parse_service_line(char **to_parse, const char *watch_for,
                              service_h * father, stype_h * type,
                              char *filename, service_h ** match,
                              service_h ** exact_match);
static int parse_opt(char **where, stype_h * type, service_h * srv);

static int set_parser(s_entry * type, char **value, char *va,
                      service_h * from_service);
static int string_parser(s_entry * type, char **value, char *va,
                         service_h * from_service);
static int strings_parser(s_entry * type, char **value, char *va,
                          service_h * from_service);
static int iint_parser(s_entry * type, char **value, char *va,
                       service_h * from_service);

static int is_valid(char *string);

static int is_valid(char *string)
{
    int i;

    for (i = 0; string[i] != 0; i++)
        if (string[i] <= 32)
            return (FALSE);

    return (TRUE);
}

static void err_print_line2t(char *point, const char *message, e_mt err,
                             const char *file, const char *func, int codeline)
{
    char line[101];
    char pstr[101];
    int len = 0;
    int i = 0;

    /* make sure global filename is set */
    if (!g_pointer || !g_filename)
        return;

    /* The message must be an error message to be shown */
    if (err == MSG)
        return;

    /* make sure point string is nulled */
    pstr[0] = '\0';

    /* set up a pointer string like "          ^" pointing to where we parse */
    len = (point - g_pointer);
    /* calulate the posision, the parser have come */
    if (point && (len > 1 || len < 99))
    {

        /* fill the space with blanks */
        for (i = 0; i <= len; i++)
            pstr[i] = ' ';

        /* then, put the marker. */
        pstr[i] = '^';
        pstr[i + 1] = '\0';
    }

    /* calculate lenght to newline */
    len = strcspn(g_pointer, "\n");

    if (len > 100)
        len = 100;                          /* TODO: really needed, or implicated by some constraints?? */

    /* fill the line, with a line with data */
    strncpy(line, g_pointer, len);
    line[len] = '\0';

    if (pstr[0])
        print_error(err, file, func, codeline, "%s\nfile: %s\n\"%s\"\n%s\n\n",
                    message, g_filename, line, pstr);
    else
        print_error(err, file, func, codeline, "%s\nfile: %s\n\"%s\"\n\n",
                    message, g_filename, line);

}



/* Load a service from a service_to_find or process_path */
service_h *initng_i_parser(const char *service_to_find)
{
    char filetoparse[200];
    service_h *got_serv = NULL;
    char *path = NULL;
    const char *service = NULL;

    assert(service_to_find);

    /* Make sure the filename of 200 chars abow are more than enought */
    if (strlen(service_to_find) > 50)
    {
        F_("Service name to long, initng_i_parser can't look for this service!\n");
        return (NULL);
    }

    /*
     * now set up path and service.
     * if service_to_find is system/initial/udevd
     * path should be "system/initial"
     * and service "udevd"
     */

    /* set up a patch copy, has to be free() */
    path = st_get_path(service_to_find);

    /* this is a pointer to a entry in service_to_find */
    service = st_strip_path(service_to_find);

    D_("initng_i_parser(%s);\n", service_to_find);
    /* example  INITNG_ROOT=/etc/initng service_to_find="deamon/sshd INITNG_EXT=".i" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.i" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.i" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.i" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.i */
    /* TRY NO 6, try load   "/etc/initng/sshd.i" */
    /* TRY NO 7, try load   "/etc/initng/daemon.i */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/");
    strcat(filetoparse, service);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service_to_find);
    strcat(filetoparse, "/default" INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 4, try "/etc/initng/daemon/default.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, "/default" INITNG_EXT);

    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, "/");
    strcat(filetoparse, path);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 6, try "/etc/initng/sshd.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, service);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    /* TRY NO 7, try "/etc/initng/daemon.i" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, path);
    strcat(filetoparse, INITNG_EXT);
    if ((got_serv = parse_file(filetoparse, service_to_find)))
    {
        free(path);
        return (got_serv);
    }
    filetoparse[0] = '\0';

    D_("Was not able to parse: %s\n", service_to_find);
    free(path);
    return (NULL);
}

/* parse a file for services */
static service_h *parse_file(char *filename, const char *watch_for)
{
    /* Pointer to content read, and pointer to where we are reading */
    char *file_content = NULL;
    char *file = NULL;


    /* if service is matching, it is set, if service is exeact_match this is set */
    service_h *match = NULL;
    service_h *exact_match = NULL;

    /* standard assertions */
    assert(filename);
    assert(watch_for);

    /* set the error message filename */
    g_filename = filename;

    D_("parse_file(%s);\n", filename);

    /* read config file */
    if (!open_read_close(filename, &file_content))
    {
        D_("Can't open config file %s.\n", filename);
        return (NULL);
    }
    file = file_content;


    /* main parse line by line loop */
    while (file[0] != '\0')
    {
        g_pointer = file;
        stype_h *t, *y = NULL;

        /* skip leading spaces */
        JUMP_SPACES(file);
        if (file[0] == '\0')
            break;

        /* skip lines starting with '#' */
        if (file[0] == '#' || file[0] == '\n' || file[0] == ';')
        {
            REALLY_JUMP_TO_NEXT_ROW(file);
            continue;
        }
        /*
         * parsometer:
         *   service test {
         *   ^
         */

        /* try to find the service type keyword */
        while_stypes(t)
        {
            if (st_cmp(&file, t->name))
            {
                y = t;
                break;
            }
        }
        if (!y)
        {
            match = NULL;
            goto free_and_return;
        }

        /* parse line, with options from that service type */
        if (parse_service_line(&file, watch_for, NULL, t,
                               filename, &match, &exact_match) == TRUE)
        {
            /* if we found RIGHT what we wanted */
            if (exact_match)
            {
                match = exact_match;
                goto free_and_return;
            }

            JUMP_TO_NEXT_ROW(file);
            continue;
        }

        /* else BAIL OUT */
        match = NULL;
        goto free_and_return;
    }                                       /* end while */

  free_and_return:
    if (file_content)
    {
        free(file_content);
        file_content = NULL;
    }

    /* Return what we got, and hope for the best */
    return (match);
}




static int parse_service_line(char **to_parse, const char *watch_for,
                              service_h * father, stype_h * type,
                              char *filename, service_h ** match,
                              service_h ** exact_match)
{
    char *name = NULL;
    char *father_name = NULL;
    service_h *new_service = NULL;
    service_h *duplicate = NULL;
    int len;

    assert(to_parse);
    assert(*to_parse);
    assert(watch_for);

    /* for debuging */
    print_line(*to_parse, "parse_service_line");

    /* service test : class { */
    /*        |               */
    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);
    if (ENDED(*to_parse))
    {
        err_print_line(*to_parse, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* service test : class { */
    /*         |              */

    /* this fetches name, and increase *to_parse */
    if (!(name = st_dup_next_word(to_parse)))
    {
        err_print_line(*to_parse, "Did not get a name!");
        return (FALSE);
    }

    if (!is_valid(name))
    {
        err_print_line(*to_parse, "Name contains invalid characters.");
        return (FALSE);
    }

    /* service test : class { */
    /*             |          */

    D_("parse_service_line(s,%s): service name: \"%s\"\n", watch_for, name);

    /* jump forward on spaces */
    JUMP_SPACES(*to_parse);
    if (ENDED(*to_parse))
    {
        err_print_line(*to_parse, "Line ended unexpectedly.");
        free(name);
        return (FALSE);;
    }

    /* service test : class { */
    /*              |         */
    if ((*to_parse)[0] == ':')
    {
        (*to_parse)++;
        /* service test : class { */
        /*               |        */

        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);
        if (ENDED(*to_parse))
        {
            err_print_line(*to_parse, "Line ended unexpectedly.");
            free(name);
            return (FALSE);
        }
        /* service test : class { */
        /*                |       */

        if (!(father_name = st_dup_next_word(to_parse)))
        {
            err_print_line(*to_parse, "Unable to fetch fathername.");
            free(name);
            return (FALSE);
        }

        if (!is_valid(father_name))
        {
            err_print_line(*to_parse,
                           "Father name contains invalid characters.");
            return (FALSE);
        }

#ifdef DEBUG
        if (father_name)
            D_("Father is: %s, set from \"server : father\"\n", father_name);
#endif

        /* service test : class { */
        /*                     |  */
        /* jump forward on spaces */
        JUMP_SPACES(*to_parse);
        if (ENDED(*to_parse))
        {
            err_print_line(*to_parse, "Line ended unexpectedly.");
            free(name);
            free(father_name);
            return (FALSE);
        }
    }

    /* service test : class { */
    /*                      | */
    if ((*to_parse)[0] != '{')
    {
        /* something is wrong */
        err_print_line(*to_parse, "error parsing, expected something else.");
        free(name);
        if (father_name)
            free(father_name);
        return (FALSE);
    }

    /* create a service entry - if there exits an duplicate, skip to the end of section */
    if ((duplicate = service_db_find_by_exact_name(name)))
    {
        free(name);
        if (father_name)
            free(father_name);
        new_service = duplicate;
        len = chars_to_end_of_this_stack(*to_parse);
        if (len == FALSE)
            return (FALSE);

        (*to_parse) += len;
    }
    else
    {
        /* jump to first char after start tag and begin from there. */
        (*to_parse)++;
        /* service test : class { */
        /*                       | */

        new_service = service_db_new(name, type);

        /* from here, dont use name anymore, use new_service->name */
        free(name);
        name = NULL;

        /* FROM NOW, we should not free name or father_name, becouse its ADDED, and required */

        /* check so that it was allocated! */
        if (!new_service)
        {
            F_("Unable to allocate space for new service.\n");
            err_print_line(*to_parse,
                           "Unable to allocate space for new service.");
            if (father_name)
                free(father_name);
            return (FALSE);
        }

        /* set the father to the service */
        if (father_name)
        {
            if (father)
                new_service->father = father;
            new_service->father_name = father_name;
        }

        /* set type if not set. */
        if (father)
            new_service->type = type;

        /* carry on until segment stop or eof, this will handle all lines in current section */
        while ((*to_parse)[0] != '\0' && (*to_parse)[0] != '}')
        {
            stype_h *t = NULL;
            stype_h *y = NULL;

            /* upgrade line pointer that print_line_error uses */
            g_pointer = (*to_parse);
            /* skip spaces and empty lines with them */
            JUMP_SPACES(*to_parse);

            /* check that this is not the end */
            if (!(*to_parse)[0])
                break;

            /* if the row is ending */
            if ((*to_parse)[0] == '\n' || (*to_parse)[0] == ';')
            {
                /* walk to next one */
                (*to_parse)++;
                continue;
            }

            /* skip lines with '#' */
            if ((*to_parse)[0] == '#')
            {
                REALLY_JUMP_TO_NEXT_ROW(*to_parse);
                continue;
            }

            /****  Check for subclasses ****/
            /* try to find the service type keyword */
            while_stypes(t)
            {
                /* look for matching keywords */
                if (st_cmp(to_parse, t->name))
                {
                    y = t;
                    break;
                }
            }
            /* if found */
            if (y)
            {
                /* parse it */
                if (!parse_service_line(to_parse, watch_for, NULL, y,
                                        filename, match,
                                        exact_match) == FALSE)
                {
                    err_print_line(*to_parse,
                                   "Parse of service line failed!");
                    service_db_free(new_service);
                    new_service = NULL;
                    return (FALSE);
                }
                JUMP_TO_NEXT_ROW(*to_parse);
                continue;
            }

            /* parse line - search for keywords in g.option_table */
            if (parse_opt(to_parse, type, new_service))
            {
                /* ok, this line is parsed, go to next */
                JUMP_TO_NEXT_ROW(*to_parse);
                continue;
            }

            /* Bad content, coud not be parsed */
            /* free and reset */
            service_db_free(new_service);
            new_service = NULL;
            return (FALSE);
        }

        service_db_set_string(&FROM_FILE, new_service, i_strdup(filename));
        /* add to service db */
        if (!service_db_add(new_service))
        {
            F_("Could not add service to service db!\n");
            service_db_free(new_service);
            new_service = NULL;
            return (FALSE);
        }
    }

    if (strcmp(new_service->name, watch_for) == 0)
    {
        D_("Found EXACT match: %s\n", new_service->name);
        (*exact_match) = new_service;
        return (TRUE);
    }

    /* check if this is the service we are looking for. */
    if (service_match(watch_for, new_service->name) == TRUE)
    {
        D_("Found :%s\n", new_service->name);
        (*match) = new_service;
        return (TRUE);
    }

    D_("No match\n");
    /* return null if not */
    return (TRUE);
}


#define MAX_OPT_LEN 100
#define MAX_VAR_LEN 100

static int parse_opt(char **where, stype_h * type, service_h * srv)
{
    s_entry *current = NULL;
    char opt_name[MAX_OPT_LEN + 1];
    char *var_name = NULL;
    int opt_len = 0;
    int var_len = 0;
    int result = TRUE;

    print_line(*where, "parse_opt");

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "|                               "
     */

    JUMP_SPACES(*where);
    if (ENDED(*where))
    {
        err_print_line(*where, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "   |                            "
     */

    /* skip empty services */
    if ((*where)[0] == '}')
    {
        err_print_line(*where, "found a } char here!");
        return (FALSE);
    }

    /* count number of chars this option name have */
    if ((opt_len = strcspn((*where), "\n; \t=\"'")) < 1)
    {
        err_print_line(*where, "option name to short!");
        return (FALSE);
    }

    /* get option name and strip all blanks from it */
    if (opt_len >= MAX_OPT_LEN)
    {
        err_print_line(*where, "Max optlen 100 reached!");
        return (FALSE);
    }
    strncpy(opt_name, (*where), opt_len);
    opt_name[opt_len] = '\0';

    /* walk */
    (*where) += opt_len;
    if (ENDED(*where))
    {
        err_print_line(*where, "Line just ended after variable name.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "         |                      "
     */

    JUMP_SPACES(*where);
    if (ENDED(*where))
    {
        err_print_line(*where, "Line just ended after variable name.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "          |                     "
     */
    print_line(*where, "Got here:");

    if ((*where)[0] != '=' && (*where)[0] != ';')
    {
        /* count number of chars this option name have */
        if (!(var_len = strcspn((*where), "\n; \t=\"'")))
        {
            err_print_line(*where, "variable name to short!");
            return (FALSE);
        }

        /* get option name and strip all blanks from it */
        if (var_len >= MAX_VAR_LEN)
        {
            err_print_line(*where, "Maximum variable name of 100 met.");
            return (FALSE);
        }

        /* copy the string */
        var_name = i_strndup(*where, var_len);

        if (!is_valid(var_name))
        {
            err_print_line(*where,
                           "Variable name contains invalid characters.");
            return (FALSE);
        }

        (*where) += var_len;
        /* Parsometer
         *
         * "   script start = /usr/bin/gdm; "
         * "               |                "
         */
        JUMP_SPACES(*where);
        if (ENDED(*where))
        {
            err_print_line(*where, "Line ended unexpectedly.");
            free(var_name);
            return (FALSE);
        }
        /* Parsometer
         *
         * "   script start = /usr/bin/gdm; "
         * "                |               "
         */

    }

    /* walk the option db */
    while_option_db(current)
    {
        /* temporary pointer storage */
        s_entry *tmp = current;

        /* make sure opt_name is set */
        if (!current->opt_name)
            continue;

        /* check that option is not marked for a sertain service type */
        if (current->ot && current->ot != type)
            continue;

        /* check string lenght of word, this is faster then strcasecmp */
        if (opt_len != current->str_len)
        {
            /*F_("Lenght is not correct.\n"); */
            continue;
        }

        /* finaly, check if opt_name is what we are looking for */
        if (strncmp(current->opt_name, opt_name, opt_len) != 0)
            continue;

        D_("parse_opt(%s): option \"%s\" value at option.\n",
           srv->name, current->opt_name);

        /* if this is a alias, browse forward to the correct opt_type */
        while (tmp->opt_type == ALIAS && tmp->alias)
            tmp = tmp->alias;

        D_("option type %i, var_name: %s, opt_name: %s\n", tmp->opt_type,
           var_name ? var_name : "(none)", current->opt_name);

        /* now switch the opt_type */
        switch (tmp->opt_type)
        {
            case STRING:
            case VARIABLE_STRING:
                result = string_parser(current, where, var_name, srv);
                break;
            case STRINGS:
            case VARIABLE_STRINGS:
                result = strings_parser(current, where, var_name, srv);
                break;
            case SET:
            case VARIABLE_SET:
                result = set_parser(current, where, var_name, srv);
                break;
            case INT:
            case VARIABLE_INT:
                result = iint_parser(current, where, var_name, srv);
                break;
            default:
                err_print_line(*where,
                               "Unknown, or undefined type to parse!");
                result = FALSE;
                break;
        }
    }

    if (result == FALSE)
    {
        /* bail out if we did not find a match */
        err_print_line(*where, "Did not get a MATCH!");
    }

    return (result);
}

static int set_parser(s_entry * type, char **value
                      __attribute__ ((unused)), char *va,
                      service_h * from_service)
{
    assert(from_service);
    assert(*value);
    assert(value);

    /*DEBUG{
       char print[200];
       sprintf(print, "set parser: type:%s va: %s service: %s", type->opt_name, va, from_service->name);
       err_print_line(*value, print);
       } */

    /* Parsometer
     *
     * "   enable;"
     * "         |"
     */

    if ((*value)[0] == ';')
    {
        D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
        service_db_set_var(type, va, from_service);
        return (TRUE);
    }

    /* Parsometer
     *
     * "   enable = yes"
     * "          |    "
     */
    if ((*value)[0] != '=')
    {
        err_print_line(*value, "There should be an ';' or '=' here!");
        return (FALSE);
    }

    /* skip the '=' char */
    (*value)++;
    JUMP_SPACES(*value);
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   enable = yes"
     * "            |  "
     */
    if (strncmp(*value, "yes", 3) == 0 || strncmp(*value, "true", 4) == 0)
    {
        D_("set_parser(%s,s,%s);\n", type->opt_name, from_service->name);
        service_db_set_var(type, va, from_service);
        return (TRUE);
    }

    /* else, remove */
    service_db_remove_var(type, va, from_service);
    return (TRUE);
}


static int string_parser(s_entry * type, char **value, char *va,
                         service_h * from_service)
{
    char *to;
    int len = 0;

    assert(from_service);
    assert(*value);
    assert(value);

    /*err_print_line(*value, "stringparser"); */
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                |               "
     */

    if ((*value)[0] != '=')
    {
        err_print_line(*value, "There should be an = here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                 |              "
     */

    JUMP_SPACES(*value);
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = { /usr/bin/gdm };"
     * "   script start = /usr/bin/gdm; "
     * "                  |             "
     */

    /* get the lenght */
    if ((*value)[0] == '{')
    {
        len = chars_to_end_of_this_stack(*value);
        /* strip trailing brackets */
        if (len != FALSE)
        {
            (*value)++;
            len--;
        }

    }
    else
    {
        len = strcspn(*value, "\n;");
    }

    if (len == FALSE)
    {
        err_print_line(*value, "Bracket count mistmatch.");
        return (FALSE);
    }

    /* make sure the end is a ';' */
    if ((*value)[len] != '}' && (*value)[len] != ';')
    {
        err_print_line(*value, "There should be an ; here.");
        if ((*value)[len] == '\n')
            F_("Its a *newline* instead.\n");
        else
            F_("Its a %c instead.\n", (*value)[len]);
        return (FALSE);
    }

    /* ok, copy the code */
    if (!(to = i_strndup(*value, len)))
    {
        err_print_line(*value, "Failed copying.");
        return (FALSE);
    }

    /* walk */
    (*value) += len;

    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        free(to);
        return (FALSE);
    }
    /* Parsometer
     *
     * "   script start = /usr/bin/gdm; "
     * "                              | "
     */

    service_db_set_string_var(type, va, from_service, to);
    return (TRUE);
}

static int iint_parser(s_entry * type, char **value, char *va,
                       service_h * from_service)
{
    char to[11];
    int len = 0;

    assert(from_service);
    assert(*value);
    assert(value);

    /*err_print_line(*value, "intparser"); */
    /* Parsometer
     *
     * "   pause = 10; "
     * "         |     "
     */

    if ((*value)[0] != '=')
    {
        err_print_line(*value, "Missing an '=' here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   pause = 10; "
     * "          |    "
     */

    JUMP_SPACES(*value);
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   pause = 10; "
     * "           |   "
     */

    len = strcspn(*value, "\n;");
    /* make sure the end is a ';' */
    if ((*value)[len] != ';')
    {
        err_print_line(*value, "Missing an ';' char.");
        return (FALSE);
    }

    /* make sure value is not to big */
    if (len >= 10)
    {
        err_print_line(*value, "Length is bigger than 10.");
        return (FALSE);
    }

    /* ok, copy the code */
    strncpy(to, *value, len);
    to[len] = '\0';

    /* walk */
    (*value) += len;
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    service_db_set_int_var(type, va, from_service, atoi(to));
    return (TRUE);
}

static int strings_parser(s_entry * type, char **value, char *va,
                          service_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    assert(*value);
    assert(value);

    /*err_print_line(*value, "stringS parser"); */
    /* Parsometer
     *
     * "   need = test service; "
     * "        |               "
     */

    if ((*value)[0] != '=')
    {
        err_print_line(*value, "Missing an '=' here.");
        return (FALSE);
    }

    (*value)++;
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }

    /* Parsometer
     *
     * "   need = test service; "
     * "         |              "
     */
    JUMP_SPACES(*value);
    if (ENDED(*value))
    {
        err_print_line(*value, "Line ended unexpectedly.");
        return (FALSE);
    }
    /* Parsometer
     *
     * "   need = test service; "
     * "          |             "
     */

    while ((to = st_dup_next_word(value)))
    {
        service_db_set_another_string_var(type, va, from_service, to);
        JUMP_SPACES(*value);
    }
    /* Parsometer
     *
     * "   need = test service; "
     * "                      | "
     */

    if ((*value)[0] != ';')
    {
        err_print_line(*value, "End char is not an ';'.");
        return (FALSE);
    }
    return (TRUE);
}

int module_init(const char *version)
{

    D_("i_parser: module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    return (initng_add_hook(PARSERS, 50, &initng_i_parser));

}

void module_unload(void)
{
    D_("i_parser: module_unload();\n");
    initng_del_hook(PARSERS, &initng_i_parser);
}
