/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <assert.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_static_states.h"
#include "../../src/initng_service.h"

#include "initng_last.h"

s_entry LAST = { "last", SET, 4, NULL,
    "If this option is set, you will be sure this service is started last."
};


/* returns TRUE if all use deps are started */
static int check_last(active_h * service, a_state_h * status)
{
    active_h *current = NULL;

    assert(service);
    assert(service->name);

    /* if we are trying to set status START_DEP_MET */
    if (status != &START_DEP_MET)
        return (TRUE);

    /* And LAST is set. */
    if (!active_db_is(&LAST, service))
        return (TRUE);

    /* ok check with all service */
    D_("LAST: walking through service db\n");
    while_active_db(current)
    {
        /* dont check ourself */
        if (current == service)
            continue;

        /* ignore runlevels */
        /* TODO 20051105 SaTaN0rX: i do not know why i need to check for this */
        if (current->type == &TYPE_RUNLEVEL)
            continue;

        /* if this service also should be started last, continue */
        if (active_db_is(&LAST, current))
            continue;

        /* TODO, use dep_on_deep */
        /* if current need service that should be last */
        if (active_db_dep_on_deep(current, service) == TRUE)
        {
            /* dont wait, becouse this wait is circular */
            D_("Service %s depends on %s\n", service->name, current->name);
            continue;
        }

        if (IS_STARTING(current))
        {
            D_("Service %s is also starting, and %s should be started last\n",
               current->name, service->name);
            return (FALSE);
        }
    }

    /* ok, filaly last start this */
    return (TRUE);
}


int module_init(const char *version)
{
    D_("module_init();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_sdt_add(&LAST);
    initng_add_hook(CONFIRM_ASTATUS_CHANGE, 80, &check_last);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_sdt_del(&LAST);
    initng_del_hook(CONFIRM_ASTATUS_CHANGE, &check_last);
}
