
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include <errno.h>

/* #include "ngc.h" */
#include "../../src/initng.h"
#include "../../src/initng_control_command.h"
#include "../../src/initng_active_state.h"
#include "../../initng-paths.h"

#include "initng_ngc2.h"

/* maximum size of string rescived, warning ngc -p can contain a lot */
#define MAX_RESCIVE_SIZE 100000

#define TRUE 1
#define FALSE 0
int debug = FALSE;

int header_printed = FALSE;

/* Global comunication socket */
static int sock = -1;

/* functions */
void print_usage(void);
int open_socket(void);
int close_socket(void);
static result_desc *send_command(const char c, const char *l, const char *o);
void command_fetch_string(void);
int command_fetch_int(void);
int main(int argc, char *argv[]);


int ngc_hlp(void);
int ngc_active_db(void);
int ngc_option_db(void);

void bad_syntax(void);
char *Argv;


#define FILL_SPACES(string, no) { \
    int t=0; \
    while(t<no) \
    { \
	if((string)[t]=='\0') \
	{ \
	    (string)[t]=' '; \
	    (string)[t+1]='\0'; \
	} \
	t++; \
    } \
}

/* print usage on screen */
void print_usage(void)
{
    send_command('h', NULL, NULL);
}

/* open socket to communicate with */
int open_socket()
{
    int len;
    struct sockaddr_un sockname;

    /* Create the socket. */
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0)
    {
        printf("Failed to init socket\n");
        return (FALSE);
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;
    if (debug)
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_TEST);
    else
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_REAL);

    len += sizeof(sockname.sun_family);

    if (connect(sock, (struct sockaddr *) &sockname, len) < 0)
    {
        close(sock);
        printf("Error connecting to initng socket\n");
        printf("This probably means that you're not running initng.\n");
        exit(1);
        return (FALSE);
    }
    return (TRUE);
}


/* close pipes */
int close_socket(void)
{
    if (sock != -1)
    {
        close(sock);
        sock = -1;
    }
    return (sock);
}



/* send a command */
static result_desc *send_command(const char c, const char *l, const char *o)
{

    read_header header;
    result_desc *result = NULL;

    /* clear structure just in case */
    memset(&header, 0, sizeof(header));
    header.p_ver = PROTOCOL_VERSION;

    /* fill the header with data */
    header.c = c;
    if (l)
        strncpy(header.l, l, 100);
    else
        header.l[0] = '\0';
    if (o)
        strncpy(header.o, o, 200);
    else
        header.o[0] = '\0';


    /*printf("Sending: %c, %s, %s\n", c, l ,o); */
    if ((sock < 0) && (open_socket() == FALSE))
    {
        printf("Socket is not open, or failed to open!\n");
        exit(1);
    }

    if (write(sock, &header, sizeof(read_header)) <
        (signed) sizeof(read_header))
    {
        printf("Could not send command!\n");
        exit(1);
    }

    /* no idea to stress initng */
    usleep(50000);

    /* if we are reloading initng, we wont have an answ from it, so don't fetch any */
    if (header.c == 'c')
    {
        printf
            ("Can't give you feedback if reloading of initng succeeds or not.\n");
        exit(0);
    }

    /* FETCH THE RESULT */

    /* calloc the result space */
    result = calloc(1, sizeof(result_desc));

    /* read header data */
    if (TEMP_FAILURE_RETRY(read(sock, result, sizeof(result_desc))) !=
        sizeof(*result))
    {
        printf("failed to fetch the result\n");
        exit(1);
    }

    /* check that protocol matches */
    if (result->p_ver != PROTOCOL_VERSION)
    {
        printf("PROTOCOL_VERSION missmatch :\n");
        printf
            (" initng_protocol_version: %i, client_protocol_version: %i !\n",
             result->p_ver, PROTOCOL_VERSION);
        printf(" initng_version: %s, ngc version: %s\n", result->version,
               VERSION);
        exit(1);
    }

    if (header_printed == FALSE)
    {
        printf("Got an response from initng, version: %s\n", result->version);

        if (strcmp(result->version, VERSION) != 0)
        {
            printf("Version of initng and version of ngc differs!\n");
            printf
                ("this might be a bad thing, that can in worst case cause problems!\n");
        }

        /*printf("got an read: command %c, status %i, type %i\n", result->c,
           result->s, result->t); */
        header_printed = TRUE;
    }

    switch (result->s)
    {
        case S_FALSE:
            printf("\nThe result of -%c is false = %i!\n\n", result->c,
                   result->s);
            free(result);
            return (NULL);

        case S_REQUIRES_OPT:
            printf("\nThe command -%c requires an option.\n\n", result->c);
            free(result);
            return (NULL);

        case S_NOT_REQUIRES_OPT:
            printf("\nThe command -%c must have no option.\n\n", result->c);
            free(result);
            return (NULL);

        case S_INVALID_TYPE:
            printf
                ("\nThe data returning of this command ( -%c ) is an unknown type.\n\n",
                 result->c);
            free(result);
            return (NULL);

        case S_COMMAND_NOT_FOUND:
            printf("\nCommand -%c is not found.\n\n", result->c);
            free(result);
            return (NULL);

        case S_TRUE:
            return (result);
    }
    return (NULL);
}

/* return the fetched string */
void command_fetch_string(void)
{
    char tmp_read[MAX_RESCIVE_SIZE + 1];
    ssize_t len;

    while ((len =
            TEMP_FAILURE_RETRY(read(sock, &tmp_read, MAX_RESCIVE_SIZE))) > 0)
    {
        tmp_read[len] = '\0';
        printf("%s", tmp_read);
    }

    /* TODO: error handling??? */

    close_socket();
}

/* return the fetched return number */
int command_fetch_int(void)
{
    char got_from_read[20];

    got_from_read[read(sock, &got_from_read, 19)] = '\0';
    close_socket();
    return (atoi(got_from_read));
}

static int ngc_start_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;

    printf("Starting service: \n");

    while (counter < 30)
    {
        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            printf("\tFailed to fetch result\n");
            close_socket();
            return FALSE;
        }

        close_socket();
        /*printf("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */
        if (row.i == IS_UP || row.i == IS_WAITING || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*printf("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);

        /* make sure we got one */
        if (!result || result->c != 's')
        {
            printf("\tInvalid result!\n");
            close_socket();
            return (FALSE);
        }

        counter++;
    }
    close_socket();

    if (row.i == IS_UP || row.i == IS_WAITING)
    {
        printf("\tService %s started sucessfully! (%s)\n", row.n, row.s);
        return (TRUE);
    }

    if (counter == 10)
    {
        printf("Timeout on start service, might be started soon anyway.\n");
        return (FALSE);
    }


    printf("\tFailed to start service, %s (%s)\n", row.n, row.s);
    return (FALSE);
}


static int ngc_stop_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;

    printf("Stopping service: \n");

    if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
        sizeof(row))
    {
        printf("failed to read response\n");
        close_socket();
        return FALSE;
    }

    close_socket();


    while (counter < 30)
    {
        /*printf("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */

        if (row.i == IS_DOWN || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*printf("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);


        /* make sure we got one */
        if (!result || result->c != 's')
        {
            printf("\tInvalid result!\n");
            close_socket();
            return (FALSE);
        }
        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            printf("failed to read response\n");
            close_socket();
            return FALSE;
        }

        close_socket();

        counter++;
    }
    close_socket();

    if (row.i == IS_DOWN)
    {
        printf("\tService stopped sucessfully!\n");
        return (TRUE);
    }

    if (counter == 10)
    {
        printf("Timeout on stop service, might be stopped soon anyway.\n");
        return (FALSE);
    }

    printf("\tFailed to stop service, %s (%s)\n", row.n, row.s);
    return (FALSE);
}


static int handle_result(result_desc * result)
{
    /* special handleing commands */
    switch (result->c)
    {
        case 'h':
            return (ngc_hlp());
        case 's':
            return (ngc_active_db());
        case 'O':
            return (ngc_option_db());
        case 'l':
            return (ngc_active_db());
        case 'u':
            return (ngc_start_result(result));
        case 'd':
            return (ngc_stop_result(result));
    }

    /* handle by type */
    switch (result->t)
    {
        case DATA_COMMAND:
            printf("Uncatched data command!\n");
            bad_syntax();
            exit(1);

        case TRUE_OR_FALSE_COMMAND:
            {
                int r = command_fetch_int();

                if (r == TRUE)
                    printf(" command -%c returned SUCCESS! \n", result->c);
                else
                    printf(" command -%c returned FAILURE! \n", result->c);

                return (r);
            }
        case INT_COMMAND:
            printf("INT_COMMAND_RESULT: %i\n", command_fetch_int());
            return (TRUE);
        case PRINT_COMMAND:
        case STRING_COMMAND:
            command_fetch_string();
            return (TRUE);
        case VOID_COMMAND:
            printf("VOID_COMMAND\n");
            return (TRUE);
        case COMMAND_FAIL:
        default:
            close_socket();
            printf("unkown data!, no haldler!\n");
            bad_syntax();
            exit(1);

    }
    printf("Error, should not got here!\n");
    exit(22);
}

/* THIS IS MAIN */
int main(int argc, char *argv[])
{
    int cc = 1;
    result_desc *result = NULL;

    header_printed = FALSE;
    Argv = argv[0];

    /* print banner */
    printf("Next Generation init Control. version: %s\n", VERSION);
    printf("written by Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");



    /* check if cmd-line contains "ngdc", if set debug */
    if (strstr(argv[0], "ngdc"))
    {
        printf("Warning. This is ngdc!\n");
        debug = TRUE;
    }

    if (debug == FALSE && getuid() != 0)
    {
        printf("You need root access to communicate with initng.\n");
        exit(2);
    }

    /* make sure there are any arguments at all */
    if (argc < 2)
        print_usage();


    while (argv[cc])
    {
        char *opt = NULL;

        /* every fresh start needs a '-' char */
        if (argv[cc][0] != '-')
        {
            bad_syntax();
            exit(1);
        }

        /* check that there is a char after the '-' */
        if (!argv[cc][1])
        {
            bad_syntax();
            exit(1);
        }

        /* if next option contains data, but not starting with an '-', its considered an option */
        if (argv[cc + 1] && argv[cc + 1][0] != '-')
            opt = argv[cc + 1];


        /* if it is an --option */
        if (argv[cc][1] == '-')
        {
            if (!(result = send_command('\0', &argv[cc][2], opt)))
            {
                printf("Command %s, %s - Unable to send command!\n",
                       &argv[cc][2], opt);
                exit(1);
            }
            if (!handle_result(result))
            {
                printf("Command %s, %s - Bad result!\n", &argv[cc][2], opt);
                exit(1);
            }
            if (result)
                free(result);
            /* if it is an -o option */
        }
        else
        {
            if (!(result = send_command(argv[cc][1], NULL, opt)))
            {
                printf("Command %c, %s - Unable to send command!\n",
                       argv[cc][1], opt);
                exit(1);
            }
            if (!handle_result(result))
            {
                printf("Command %s, %s - Bad result!\n", &argv[cc][1], opt);
                exit(1);
            }
            if (result)
                free(result);
        }

        cc++;
        if (opt)
            cc++;


    }
    exit(0);
}




/* get help */
int ngc_hlp(void)
{
    /* data got */
    help_row row;

    /* print head */
    printf(" ngc understand this commands:\n");
    printf(" option                          : description\n");
    printf(" ----------------------------------------------------------\n");

    while (read(sock, &row, sizeof(help_row)))

    {
        int i;
        char lname[32];

        /* copy name to the new static array */
        strncpy(lname, row.l, 25);
        lname[25] = '\0';

        switch (row.o)
        {
            case USES_OPT:
                strcat(lname, " <opt>");
                break;
            case REQUIRES_OPT:
                strcat(lname, " opt");
                break;
            default:
                break;
        }

        /* Fill out 40 blanks for nice output */
        for (i = strlen(lname); i <= 31; i++)
            lname[i] = ' ';

        lname[31] = '\0';
        if (row.c != '\0')
            printf(" [-%c] --%s: %s\n", row.c, lname, row.d);
        else
            printf("       --%s: %s\n", lname, row.d);
    }
    return (TRUE);
}


int ngc_active_db(void)
{
    active_row row;

    /* print head */
    printf(" hh:mm:ss service                                  : status\n");
    printf(" ----------------------------------------------------------\n");

    while (read(sock, &row, sizeof(active_row)))
    {
        struct tm *ts = localtime(&row.t);

        FILL_SPACES(row.n, 40);

        printf(" %.2i:%.2i:%.2i %s : %s\n", ts->tm_hour, ts->tm_min,
               ts->tm_sec, row.n, row.s);

    }
    close_socket();
    return (TRUE);
}


int ngc_option_db(void)
{
    option_row row;
    char ct[20];

    /* print head */
    printf(" %-10s %-24s = %-8s %s\n", "WhereToUSE", "Name", "ValueToUSE",
           "Description");
    printf
        (" ----------------------------------------------------------------\n");

    while (read(sock, &row, sizeof(option_row)))
    {
        switch (row.t)
        {
            case STRING:
                strcpy(ct, "STRING");
                break;
            case STRINGS:
                strcpy(ct, "STRINGS");
                break;
            case SET:
                strcpy(ct, "SET");
                break;
            case INT:
                strcpy(ct, "INT");
                break;
            case ALIAS:
                strcpy(ct, "ALIAS");
                break;
            default:
                strcpy(ct, "U_D_T");
                break;
        }

        printf(" %-10s %-24s = %-8s %s\n", row.o, row.n, ct, row.d);

    }
    close_socket();
    return (TRUE);
}


/* Avoid using goto */
void bad_syntax(void)
{
    printf("Bad syntax!\n\n");
    print_usage();
}
