
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <errno.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_active_state.h"
#include "../../src/initng_process_db.h"
#include "../../initng-paths.h"

#include "initng_reload.h"

#define SAVE_FILE      VARDIR "/initng_db_backup.v11"
#define SAVE_FILE_FAKE VARDIR "/initng_db_backup.v11"

s_command FAST_RELOAD = { 'c', "hot_reload", VOID_COMMAND, NO_OPT,
    {(void *) &cmd_fast_reload},
    "Fast Reload"
};

static void write_file(const char *filename);
static void read_file(const char *filename);

const char *module_needs[] = {
    "rlparser",
    "iparser",
    NULL
};

static void read_file(const char *filename)
{
    FILE *fil;
    active_h *new_entry = NULL;
    data_save_struct entry;
    s_data *d;
    int i;
    int pnr = 0;
    process_h *process = NULL;

    fil = fopen(filename, "r");

    if (!fil)
        return;

    while (!feof(fil))
    {
        if (!fread(&entry, sizeof(entry), 1, fil))
            continue;

        if (active_db_find_by_name(entry.name))
        {
            W_("Entry exists, won't create it!\n");
            continue;
        }

        if (!(new_entry = active_db_new(i_strdup(entry.name))))
        {
            F_("Can't create new active!\n");
            continue;
        }

        new_entry->a_state = initng_find_state(entry.state);
        if (!new_entry->a_state)
        {
            F_("Could not find a proper state to set.\n");
            continue;
        }
        {
            stype_h *stype = NULL;

            while_stypes(stype)
            {
                if (strcmp(stype->name, entry.type))
                    new_entry->type = stype;
            }
        }
        new_entry->time_got_status = entry.time_got_status;

        /* walk thru all processes */
        pnr = 0;
        while (entry.process[pnr].ptype[0] && pnr < MAX_PROCESSES)
        {
            ptype_h *pt = NULL;

            while_ptypes(pt)
            {
                if (strcmp(entry.process[pnr].ptype, pt->name) == 0)
                    break;
            }

            /* check so it was found */
            if (strcmp(entry.process[pnr].ptype, pt->name) != 0)
                continue;

            /* allocate the process */
            process = process_db_new(pt);
            if (!process)
                continue;

            /* fill the data */
            process->pid = entry.process[pnr].pid;
            process->out_pipe[0] = entry.process[pnr].stdout1;
            process->out_pipe[1] = entry.process[pnr].stdout2;
            process->r_code = entry.process[pnr].rcode;

            /* allocate a new output buffer for this process */
            process->buffer_pos = 0;
            process->buffer = NULL;

            /* add this process to the list */
            list_add(&process->list, &new_entry->processes.list);

            D_("Added process type %i to %s\n", process->pt, new_entry->name);

            pnr++;
        }

        i = 0;
        while (entry.data[i].opt_type)
        {
            d = (s_data *) initng_calloc(1, sizeof(s_data));
            d->type = initng_sdt_find(entry.data[i].type);
            if (!d->type)
            {
                F_("Did not found %s!\n", entry.data[i].type);
                free(d);
                i++;
                continue;
            }
            switch (d->type->opt_type)
            {
                case STRING:
                case STRINGS:
                case VARIABLE_STRING:
                case VARIABLE_STRINGS:
                    d->t.s = i_strdup(entry.data[i].t.s);
                    break;
                case INT:
                case VARIABLE_INT:
                    d->t.i = entry.data[i].t.i;
                    break;
                default:
                    break;
            }
            list_add(&d->list, &new_entry->data.list);
            i++;
        }
        if (active_db_add(new_entry) != TRUE)
        {
            F_("Could not add entry!\n");
            active_db_free(new_entry);
            continue;
        }

        if (get_service(new_entry) == TRUE)
            new_entry->from_service =
                service_db_find_by_name(new_entry->name);
    }

    fclose(fil);
    if (unlink(filename) != 0)
    {
        W_("Failed removing file %s !!!\n", filename);
        return;
    }
}
static void write_file(const char *filename)
{
    FILE *fil;
    active_h *current, *q = NULL;
    data_save_struct entry;
    process_h *process = NULL;
    int i;
    int pnr = 0;
    s_data *c_d = NULL;

    fil = fopen(filename, "w+");
    if (!fil)
    {
        F_("Could not open '%s' for writing\n", filename);
        return;
    }

    /* walk the active_db */
    while_active_db_safe(current, q)
    {
        if (!current->a_state)
        {
            F_("State is not set, wont save this one!\n");
            continue;
        }

        if (strlen(current->name) >= MAX_SERVICE_NAME_STRING_LEN)
        {
            F_("Service name is to long, it won't fit the fileformat spec! max is %i, won't save this service!\n", MAX_SERVICE_NAME_STRING_LEN);
            continue;
        }

        memset(&entry, 0, sizeof entry);
        strncpy(entry.name, current->name, MAX_SERVICE_NAME_STRING_LEN);
        strncpy(entry.state, current->a_state->state_name, 100);
        strncpy(entry.type, current->type->name, 100);
        entry.time_got_status = current->time_got_status;

        /* collet some proceses */
        process = NULL;
        pnr = 0;
        while_processes(process, current)
        {
            strncpy(entry.process[pnr].ptype, process->pt->name,
                    MAX_PTYPE_STRING_LEN);
            entry.process[pnr].pid = process->pid;
            entry.process[pnr].stdout1 = process->out_pipe[0];
            entry.process[pnr].stdout2 = process->out_pipe[1];
            entry.process[pnr].rcode = process->r_code;
            pnr++;
            if (pnr >= MAX_PROCESSES)
                break;
        }
        entry.process[pnr].ptype[0] = '\0';

        /* reset data */
        for (i = 0; i < MAX_ENTRYS_FOR_SERVICE; i++)
        {
            entry.data[i].opt_type = 0;
            entry.data[i].type[0] = '\0';
            entry.data[i].t.i = 0;
        }

        i = 0;
        list_for_each_entry(c_d, &current->data.list, list)
        {
            if (!c_d->type)
                continue;
            if (!c_d->type->opt_name)
                continue;

            entry.data[i].opt_type = c_d->type->opt_type;
            strncpy(entry.data[i].type, c_d->type->opt_name,
                    MAX_TYPE_STRING_LEN);

            /* copy the data */
            switch (c_d->type->opt_type)
            {
                case STRING:
                case STRINGS:
                case VARIABLE_STRING:
                case VARIABLE_STRINGS:
                    strncpy(entry.data[i].t.s, c_d->t.s, MAX_DATA_STRING_LEN);
                    break;
                case INT:
                case VARIABLE_INT:
                    entry.data[i].t.i = c_d->t.i;
                    break;
                default:
                    break;
            }

            /* save in next data entry row */
            i++;

            /* maximum data entrys that can be saved */
            if (i == MAX_ENTRYS_FOR_SERVICE)
            {
                F_("Maximum 20 data entries / service can't be saved!\n");
                break;
            }
        }

        D_("Saving : %s\n", entry.name);
        if (fwrite(&entry, sizeof(entry), 1, fil) != 1)
        {
            F_("failed to write entry '%s': %m\n", entry.name);
            /* TODO: database recovery?? */
            break;
        }
    }

    fclose(fil);

}

static void cmd_fast_reload(char *arg, FILE * fd)
{
    (void) arg;
    char *new_argv[4];

    if (g.i_am_init)
    {
        write_file(SAVE_FILE);
    }
    else
    {
        write_file(SAVE_FILE_FAKE);
        _exit(0);
    }

    new_argv[0] = i_strdup("/sbin/initng");
    new_argv[1] = i_strdup("--hot_reload");
    new_argv[2] = NULL;

    execve("/sbin/initng", new_argv, environ);
    F_("Failed to reload initng!\n");

}

/* Save a reload file for backup if initng segfaults */
static void save_backup(h_sys_state state)
{
    /* only save when system is getting up */
    if (state == STATE_UP)
    {
        /* save file */
        if (g.i_am_init)
        {
            write_file(SAVE_FILE);
        }
        else
        {
            write_file(SAVE_FILE_FAKE);
        }
        return;
    }

    /* if system is stopping, remove the SAVE_FILE */
    if (state == STATE_STOPPING)
    {
        if (g.i_am_init)
        {
            unlink(SAVE_FILE);
        }
        else
        {
            unlink(SAVE_FILE_FAKE);
        }
        return;
    }
}


int module_init(const char *version)
{
    D_("module_init(reload);\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    if (g.hot_reload)
    {
        if (g.i_am_init)
        {
            read_file(SAVE_FILE);
        }
        else
        {
            read_file(SAVE_FILE_FAKE);
        }
    }

    initng_command_add(&FAST_RELOAD);
    initng_add_hook(SWATCHERS, 90, &save_backup);
    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");
    initng_command_del(&FAST_RELOAD);
    initng_del_hook(SWATCHERS, &save_backup);
}
