
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>
#include <assert.h>
#include <errno.h>

#include "../../src/initng_active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_execute.h"
#include "../../src/initng_string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_common.h"
#include "../../src/initng_signal.h"
#include "../../src/initng_fork.h"
#include "../../src/initng_env_variable.h"

#include "initng_simple_launcher.h"

s_entry EXEC = { "exec", VARIABLE_STRING, 4, NULL,
    "The path and name for an executable."
};
s_entry EXEC_ARGS = { "exec_args", VARIABLE_STRING, 9, NULL,
    "The arguments for the executable."
};

static int simple_exec(process_h * p, active_h * s, size_t c, char **v);

#ifdef DEBUG
static void D_argv(const char *o, char **argv)
{
    int i;

    if (!argv)
        return;

    for (i = 0; argv[i]; i++)
        D_("%s[%-2i]: %s\n", o, i, argv[i]);
}
#endif

/*
 * Searches for exec in PATH.
 * 
 * Completely rewritten function.
 * Contains old Code from SaTaN0rX and DEac-.
 * @author TheLich
 * @param exec filename, which searched.
 * @return executeable file with absolute path.\nIf exec was absolute, this will be a pointer to exec.
 * On failure, it will return NULL.
 */

static char *expand_exec(char *exec)
{
    char *filename = NULL;
    size_t exec_len = 0;
    size_t path_c, i, len = 0;
    char *PATH = NULL;
    char **path_argv = NULL;
    struct stat test;


    if (!exec)
        return NULL;

    /* 11/20/2005 SaTaN0rX: preliminary support for searching the PATH
     * only search the PATH if this is NOT already an absolute path
     */

    /* if path provided, youst return */
    if (exec[0] == '/')
        return (NULL);

    /* fet exec string lenght to later use */
    exec_len = strlen(exec);

    /* get the env-path variable */
    PATH = i_strdup(getenv("PATH"));

    D_("initng_s_launch: %s is not an absolute path, searching $PATH\n",
       exec);

    /* Make sure we got a path */
    if (!PATH)
    {
        D_("No $PATH found, using default path\n");
        PATH =
            i_strdup
            ("/bin:/sbin:/usr/bin:/usr/sbin:/usr/local/bin:/usr/local/sbin");
    }

    D_("PATH determined to be %s\n", PATH);

    /* split path by ':' char */
    path_argv = split_delim(PATH, ":", &path_c);

    /* walk the list of entrys */
    for (i = 0; path_argv[i]; i++)
    {
        /* get entry lenght */
        len = strlen(path_argv[i]);

        /* what does this do? */
        filename = (char *) initng_calloc(exec_len + len + 2, sizeof(char));
        strcpy(filename, path_argv[i]);
        if (filename[len - 1] != '/')
            strcat(filename, "/");
        strcat(filename, exec);

        /* check so that file exits, if exists leave this loop */
        if ((stat(filename, &test) != -1)
            && (test.
                st_mode & (S_IXOTH | S_IXGRP | S_IXUSR | S_ISVTX | S_ISGID |
                           S_ISUID | S_IFREG)))
            break;

        /* cleanup */
        free(filename);
        filename = NULL;
    }

    /* free */
    free(PATH);
    free(path_argv);

    /* make sure we got a filename to use */
    if (!filename)
        F_("No executable found\n");

    /* return the filename */
    return filename;
}

static int initng_s_launch(active_h * service, process_h * process)
{
    const char *exec = NULL;

    assert(service);
    assert(service->name);

    /* WE ARE EXECUTING A START FILE */
    while ((exec =
            active_db_get_string_var(&EXEC, process->pt->name, service)))
    {
        const char *exec_args = NULL;
        char *exec_t = NULL, *exec_args_t = NULL;
        size_t exec_c = 0;
        size_t exec_args_c = 0;
        char *argv0 = NULL;
        char **exec_argv = NULL;
        char **exec_args_argv = NULL;
        int i = 0;
        int result = FALSE;

        /* be aware that fix_variables() return is a malloc, and needs to be free */
        exec_t = fix_variables(exec, service);

        /* argv-entries are pointer to exec_t[x] */
        exec_argv = split_delim(exec_t, WHITESPACE, &exec_c);

        /* make sure we got something from the split */
        if (!exec_argv || !exec_argv[0])
        {
            D_("split_delim on exec returns NULL.\n");
            goto go_free;
        }

        /* if it not contains a full path */
        if (exec_argv[0][0] != '/')
        {
            argv0 = expand_exec(exec_argv[0]);
            if (!argv0)
            {
                D_("%s was not found in search path.\n", exec_argv[0]);
                goto go_free;
            }
            exec_argv[0] = argv0;
        }

        /* exec_args should be parsed at the moment, too */
        exec_args =
            active_db_get_string_var(&EXEC_ARGS, process->pt->name, service);
        if (exec_args)
        {
            /* get some fixed variables, with ${VARIABLES} fixed, be aware that this is a new malloc and needs to be free() */
            exec_args_t = fix_variables(exec_args, service);

            if (!exec_args_t)
            {
                W_("Failed to fix_variables: \"%s\"\n", exec_args);
                goto go_free;
            }

            exec_args_argv =
                split_delim(exec_args_t, WHITESPACE, &exec_args_c);
            if (!exec_args_argv || !exec_args_argv[0])
            {
                D_("split_delim exec_args returns NULL.\n");
                goto go_free;
            }

            /* what does this part do ??? */
            exec_argv =
                (char **) initng_realloc(exec_argv,
                                         (exec_c + exec_args_c +
                                          1) * sizeof(char *));
            for (i = 0; exec_args_argv[i]; i++)
                exec_argv[i + exec_c] = exec_args_argv[i];

            exec_argv[i + exec_c] = NULL;
            exec_c += exec_args_c;

            /* this have to exist */
            free(exec_args_argv);
            exec_args_argv = NULL;
        }

        /* try to execute, remember the result */
        result = simple_exec(process, service, exec_c, exec_argv);

        /* this is a fallback, that we can go to if anything fail */
      go_free:
        /* Start freeing */
        if (exec_t)
        {
            free(exec_t);
            exec_t = NULL;
        }
        if (exec_argv)
        {
            free(exec_argv);
            exec_argv = NULL;
        }
        if (argv0)
        {
            free(argv0);
            argv0 = NULL;
        }
        if (exec_args_t)
        {
            free(exec_args_t);
            exec_args_t = NULL;
        }

        /* return if sucessfull */
        if (result == TRUE)
            return (TRUE);

        /* if something failed return false directly */
        if (result == FAIL)
            return (FALSE);

        /* continue the next otherwise */
    }

    return (FALSE);
}

/* 2005-11-28 DEac-: char *e, char *ea => char **argv, size_t argc
 */
static int simple_exec(process_h * process_to_exec, active_h * s, size_t argc,
                       char **argv)
{
    /* called from inside the service directory, return the PID or 0 on error */

    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */

    assert(process_to_exec);
    assert(s);
    assert(s->name);
    assert(argc);
    assert(argv);

    if ((pid_fork = initng_fork(s, process_to_exec)) == 0)
    {

#ifdef DEBUG
        D_("FROM_FORK simple_exec(%i,%s, ...);\n", argc, argv[0]);
        D_argv("simple_exec: ", argv);
#endif

        /* FINALLY EXECUTE *//* execve replaces the running process */
        execve(argv[0], argv, new_environ(s));

        /* Will never get here if execve succeeded */
        F_("EEEEEEEEEEEEEEEEEEERRRRRRRRRRRRRRRRRRRRRRRRRRROOOOOOOOOOOOOOOOOOOOOOOOOORRRRRRRRRRRRRRR!!!!!!!!!\n");
        F_("Can't execute source %s!\n", argv[0]);

        _exit(1);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);

    if (pid_fork > 0)
    {
        process_to_exec->pid = pid_fork;
        return (TRUE);
    }

    process_to_exec->pid = 0;
    return FALSE;
    /* if to test want to lock this up util fork is done ...   waitpid(pid_fork,0,0); */
}                                           /* end fork_and_exec() */


int module_init(const char *version)
{
    D_("initng_simple_plugin: module_init();\n");

    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_add_hook(LAUNCH, 40, &initng_s_launch);
    initng_sdt_add(&EXEC);
    initng_sdt_add(&EXEC_ARGS);
    return (TRUE);
}

void module_unload(void)
{
    D_("initng_simple_plugin: module_unload();\n");

    initng_sdt_del(&EXEC);
    initng_sdt_del(&EXEC_ARGS);
    initng_del_hook(LAUNCH, &initng_s_launch);

}
