
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../../src/initng.h"
#include "../../src/initng_global.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"
#include "../../src/initng_plugin.h"
#include "../../src/initng_control_command.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_static_states.h"

#include "initng_splash.h"

/*
 * 
 * "module global" variables ...
 * is it better to pack them all to a struct?
 *
 */

enum
{
    SPLASH_NONE,
    GENSPLASH,
    SPLASHY,
    BOOTSPLASH,
};


static int old_percent = 0;
static int shutting_down = 0;
static int splash_type = SPLASH_NONE;
static char paint_cmd[200] = "repaint\n";

static const char *control_file[] = {
    NULL,
    GENSPLASH_FIFO,
    SPLASHY_FIFO,
    BOOTSPLASH_FIFO
};

/* static const char* splash_message[] = {
   NULL,
   "progress %d\nrepaint\n",
   "progress %d\n"
   }; */

/*
 *
 * We add a new command, that will install the hooks
 * this command is called by the initscript that brings up
 * splash / splashy
 *
 * i hope this will increase performance ;)
 *
 */

s_command SPLASH_CMD =
    { 'S', "enable-splash", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
    {(void *) &cmd_splash},
    "to be called from gensplash / splash init scripts"
};


static FILE *try_to_open(const char *file_name)
{
    FILE *ret = NULL;
    int fd = -1;

    if ((fd = open(file_name, O_WRONLY | O_NONBLOCK)) < 0)
    {
        D_("splash: could not open control file %s\n", file_name);
        return (NULL);
    }
    ret = fdopen(fd, "w");
    D_("SPLASH: splash control fifo %s open\n", file_name);
    return (ret);
}


int print_progress(active_h * service)
{
    /*
     * for now this only works when starting up the system
     * 10/03/05 SaTaN0rX: works also for shutdown
     */
    int percent = 0;
    FILE *fd = NULL;

    D_("SPLASH: print_progress()\n");
    if (!service)
        return (TRUE);

    /*
     * 11/15/05 SaTaN0rX:
     * /proc/splash thing is for bootsplash support.
     * bootsplash is in the kernel and needs no extra daemon.
     * due to this, we just try to open BOOTSPLASH_FIFO if splash_type == SPLASH_NONE,
     * to autodetect bootsplash. IMHO it's an overkill to add an service, just to 
     * inform this plugin about bootsplash
     */
    if (splash_type == SPLASH_NONE)
    {
        D_("splash: we are called, but splash_type == SPLASH_NONE, trying to open " BOOTSPLASH_FIFO "\n");
        fd = try_to_open(BOOTSPLASH_FIFO);
        if (!fd)
            return (TRUE);
        splash_type = BOOTSPLASH;
    }

    if (!shutting_down)
    {
        percent = active_db_percent_started();
    }
    else
    {
        percent = active_db_percent_stopped();
    }

    /*
     * for some odd reason, the percent can get lower during the boot process...
     * this sucks.
     * so here's the check
     */

    if (old_percent < percent)
    {
        old_percent = percent;

        /* open splash control file */
        /* Warn here? */

        if (!fd)
            fd = try_to_open(control_file[splash_type]);

        if (!fd)
        {
            D_("splash: no control file\n");
            return (TRUE);
        }

        if (splash_type == GENSPLASH)
        {
            char svc_name[100];
            const char *svc_status = NULL;
            int i;

            i = 0;
            if (service->name)
            {
                if (strncmp(service->name, "net/", 4) == 0)
                {
                    strncpy(svc_name, service->name, 100);
                    svc_name[99] = '\0';
                    svc_name[3] = '.';
                }
                else
                {
                    while (service->name[i] && service->name[i] != '/')
                        i++;
                    if (service->name[i])
                        i++;
                    strncpy(svc_name, service->name + i, 100);
                    svc_name[99] = '\0';
                }

                if (service->a_state == &LOADING ||
                    service->a_state == &STARTING ||
                    service->a_state == &WAITING_FOR_START_DEP)
                    svc_status = "svc_start";
                if (service->a_state == &FAIL_STARTING ||
                    service->a_state == &START_DEP_FAILED)
                    svc_status = "svc_start_failed";
                if (service->a_state == &RUNNING || service->a_state == &DONE)
                    svc_status = "svc_started";
                if (service->a_state == &STOP_DEP_FAILED ||
                    service->a_state == &FAIL_STOPPING)
                    svc_status = "svc_stop_failed";
                if (service->a_state == &STOPPED)
                    svc_status = "svc_stopped";
                if (service->a_state == &STOPPING ||
                    service->a_state == &WAITING_FOR_STOP_DEP)
                    svc_status = "svc_stop";
                if (svc_status)
                    fprintf(fd, "update_svc %s %s\n", svc_name, svc_status);
            }
            /*
             * repaint is an Expensive operation, but required
             * for the text string that says 10%, etc...
             * it is better to use paint rect blah.
             * but this would mean i must determine the framebuffer resolution
             * and pare the .cfg file.
             *
             * i'm too lazy to do this.
             *
             * better would be a solution in which the script determines the paint rect
             * command and passes this information to the plugin.
             *
             * but how should i achieve this ?
             *
             * 10/03/05 SaTaN0rX: the cmd we installed will pass the information,
             * that has been determined by the script.
             */
            fprintf(fd, "progress %d\npaint\n%s\n", (65535 * percent) / 100,
                    paint_cmd);
        }
        else
        {
            /*
             * splash_type == SPLASHY || splash_type == BOOTSPLASH
             */

            if (splash_type == SPLASHY)
                fprintf(fd, "progress %d\n", percent);
            else
                fprintf(fd, "show %d\n", (65535 * percent) / 100);

#if 0                                       /* this should be done by the just_before plugin now ... IMHO an cleaner solution */
            /*
             * see if X11 is starting
             */
            if (strstr(service->name, "daemon/gdm") ||
                strstr(service->name, "daemon/kdm") ||
                strstr(service->name, "daemon/xdm") ||
                strstr(service->name, "daemon/wdm") || percent == 100)
            {
                if (service->a_state == &LOADING ||
                    service->a_state == &STARTING ||
                    service->a_state == &RUNNING ||
                    service->a_state == &DONE || percent == 100)
                {
                    /* X11 is being started or runs, or we have reached 100% */
                    if (splash_type == SPLASHY)
                    {
                        fprintf(fd, "progress 99\n");
                        fprintf(fd, "allowchvt\n");     //do we need this?
                        fprintf(fd, "exit\n");  //terminate splasy
                        //Now we should switch to the terminal X start on -but how canwe find out
                        //which terminal it is? 
                    }
                    else
                    {
                        fprintf(fd, "show 65535\n");
                        fprintf(fd, "master\n");
                    }
                }
            }
#endif
        }
        fclose(fd);
    }
    return (TRUE);
}

#if 0
 /*
  * we now use the up plugin to do this
  * the called script then uses the -S command to
  * inform the plugin
  *
  */
static void print_system_state(h_sys_state state)
{
    if (state == STATE_UP)
    {
        if (!start_new_service_named("system/splash/stop"))
            F_("Splash stopper \"system/splash/stop\" could not be executed!\n");
    }
    else if (state == STATE_STOPPING && !shutting_down)
    {
        //start_new_service_named("system/splash/shutdown");
        old_percent = 0;
        shutting_down = 1;
    }

}
#endif

static int cmd_splash(char *arg)
{
    char *p_cmd;

    if (!arg)
        return (FALSE);

    splash_type = SPLASH_NONE;

    if (strstr(arg, "gensplash"))
    {
        splash_type = GENSPLASH;
    }
    else if (strstr(arg, "splashy"))
    {
        splash_type = SPLASHY;
    }
    else if (strstr(arg, "bootsplash"))
    {
        splash_type = BOOTSPLASH;
    }

    if (strstr(arg, "shutdown"))
    {
        shutting_down = 1;
        old_percent = 0;
        /* initng_add_hook(ASTATUS_CHANGE, 80, &print_progress); */
    }
    else if (strstr(arg, "off"))
    {
        /* initng_del_hook(ASTATUS_CHANGE, &print_progress); */
        splash_type = SPLASH_NONE;
        return (TRUE);
    }
    else
    {
        /*
         * if neither shutdown nor off are specified,
         * we guess the cmd was boot.
         * this way we can mit boot.
         */
        shutting_down = 0;
        old_percent = 0;
        /* initng_add_hook(ASTATUS_CHANGE, 80, &print_progress); */
    }

    if (splash_type == SPLASH_NONE)
    {
        D_("splash: No splash type specified\n");
        return (FALSE);
    }

    if (splash_type == GENSPLASH && (p_cmd = strstr(arg, "cmd:")))
    {
        strncpy(paint_cmd, p_cmd + 4, 199);
        paint_cmd[199] = '\0';
    }

    return (TRUE);
}


int module_init(const char *version)
{
    D_("splash: module_load();\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }
    initng_add_hook(ASTATUS_CHANGE, 80, &print_progress);
    /*   initng_add_hook(SWATCHERS, 90, &print_system_state); */
    initng_command_add(&SPLASH_CMD);


    return (TRUE);
}

void module_unload(void)
{
    D_("splash: module_unload();\n");
    /*
     * maybe this is still loaded ?
     * what can happen if we try to unload it and it has already been unloaded ?
     */
    initng_del_hook(ASTATUS_CHANGE, &print_progress);
    initng_command_del(&SPLASH_CMD);

    /*    initng_del_hook(SWATCHERS, &print_system_state); */
}
