
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Anders Aagaard <aagaande@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../../src/initng.h"
#include "../../src/initng_global.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"
#include "../../src/initng_plugin.h"
#include "../../src/initng_control_command.h"
#include "../../src/initng_static_data_id.h"
#include "../../src/initng_static_states.h"

#include "initng_usplash.h"

/*
   todo : 
   update key initscripts such as fsck with output of whats going on
   only update progress if it's been more than a 10% change?
   need to clean tty's and kill usplash when system has booted. (see /etc/init.d/usplash on ubuntu systems)
   cache usplash's fifo, we're run after initial so nothing is screwing with /dev at this point.
 */

// Buffer is used for snprintf before writing to usplash fifo.
static char printf_buffer[1024];

static void usplash_write(char *output)
{
    int len = strlen(output) + 1;       // catch \0 aswell
    int fd = open(USPLASH_FIFO, O_WRONLY | O_NONBLOCK);

    if (fd > 0)
    {
        D_("USPLASH : Writing data \"%s\" to pipe\n", output);
        if (write(fd, output, len) != len)
            W_("USPLASH : Open ok, but write failed (%s)\n", strerror(errno));
    }
    else
        D_("USPLASH : Can't write data %s to pipe, not open yet.\n", output);
    close(fd);

}

#define USPLASH_WRITE(fmt,...) { snprintf(printf_buffer,1024,fmt, ## __VA_ARGS__); usplash_write(printf_buffer); }
int print_output(active_h * service)
{
    assert(service);
    assert(service->name);
    int percent = active_db_percent_started();

    if (percent != last_percent)
    {
        USPLASH_WRITE("PROGRESS %d", percent);
        last_percent = percent;
    }
    // Theese seem to take quite a bit of resources for me, and actually slow down boot.
    // 18 seconds boot without these, 26 with.
    /*    if (IS_MARK(service,&DONE))
       USPLASH_WRITE("TEXT %s : DONE",service->name)
       else if (IS_MARK(service,&RUNNING))
       USPLASH_WRITE("TEXT %s : RUNNING",service->name)
       else if (IS_MARK(service, &STARTING))
       USPLASH_WRITE("TEXT %s : STARTING",service->name) */
    return TRUE;
}

int module_init(const char *version)
{
    D_("usplash: module_load();\n");
    last_percent = 0;
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module is compiled for \"%s\" version and initng is compiled on \"%s\" version, won't load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }
    initng_add_hook(ASTATUS_CHANGE, 80, &print_output);


    return (TRUE);
}

void module_unload(void)
{
    D_("usplash: module_unload();\n");

    /*
     * maybe this is still loaded ?
     * what can happen if we try to unload it and it has already been unloaded ?
     */
    initng_del_hook(ASTATUS_CHANGE, &print_output);

    /*    initng_del_hook(SWATCHERS, &print_system_state); */
}
