
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef PROCESS_DB_H
#define PROCESS_DB_H

#define MAX_BUFFER 1024                     /* This is the maximum size of buffer in process_h */

#include <sys/types.h>
#include <unistd.h>
#include "initng_active_db.h"
#include "initng_service_cache.h"
#include "initng_list.h"

/* this doesn't work!, it will create a circular dependency */
/* so we use a struct prototype !! */
/* this works, as we only declare pointers to this struct here. */

//struct active_type;


/* struct containing data on an running process */
typedef struct t_ptype_h ptype_h;
typedef struct t_process_h process_h;

struct t_process_h
{
    ptype_h *pt;
    pid_t pid;                  /* pid of process */
    int r_code;
    int out_pipe[2];            /* pipes of process */
    char *buffer;               /* stdout buffer */
    int buffer_pos;             /* Our current position in the buffer */

    struct list_head list;      /* this process should be in a list */
};

struct t_ptype_h
{
    /* name to be used with launcher */
    const char *name;

    /* The function to call, when the process is returning */
    void (*kill_handler) (active_h * service, process_h * process);

    /* the list of process types */
    struct list_head list;
};



/* functions for creating and freeing a process struct */
process_h *process_db_new(ptype_h * ptype);
void process_db_free(process_h * free_this);


#define while_processes(current, service) list_for_each_entry_prev(current, &service->processes.list, list)
#define while_processes_safe(current, safe, service) list_for_each_entry_prev_safe(current, safe, &service->processes.list, list)

#define while_ptypes(current) list_for_each_entry_prev(current, &g.ptypes.list, list)
#define while_ptypes_safe(current, safe) list_for_each_entry_prev_safe(current, safe, &g.ptypes.list, list)

#define ptype_add(pt) list_add(&(pt)->list, &g.ptypes.list)
#define ptype_del(pt) list_del(&(pt)->list)

#endif
