
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <errno.h>

#include "initng_global.h"

#include "initng_service.h"

#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_execute.h"
#include "initng_common.h"
#include "initng_load_module.h"
#include "initng_handler.h"
#include "initng_kill_handler.h"
#include "initng_static_data_id.h"
#include "initng_static_states.h"


static int start_DAEMON_or_SERVICE(active_h * service_to_start);
static int stop_SERVICE_or_DAEMON(active_h * service_to_stop);

stype_h TYPE_SERVICE =
    { "service", &start_DAEMON_or_SERVICE, &stop_SERVICE_or_DAEMON, NULL };
stype_h TYPE_DAEMON =
    { "daemon", &start_DAEMON_or_SERVICE, &stop_SERVICE_or_DAEMON, NULL };
stype_h TYPE_VIRTUAL =
    { "virtual", &start_DAEMON_or_SERVICE, &stop_SERVICE_or_DAEMON, NULL };
stype_h TYPE_RUNLEVEL =
    { "runlevel", &start_DAEMON_or_SERVICE, &stop_SERVICE_or_DAEMON, NULL };
stype_h TYPE_CLASS = { "class", NULL, NULL, NULL };

void initng_add_static_stypes(void)
{
    stype_add(&TYPE_SERVICE);
    stype_add(&TYPE_DAEMON);
    stype_add(&TYPE_VIRTUAL);
    stype_add(&TYPE_RUNLEVEL);
    stype_add(&TYPE_CLASS);
}


#ifdef NONOINTE
    /* it might not yet been stopped */
if (IS_STOPPING(service_to_start))
{
    /* remark it as started */
    D_("service %s did never have time to stop!\n", service_to_start->name);
    if (service_to_start->type == &TYPE_DAEMON)
        mark_service(service_to_start, &RUNNING);
    else
        mark_service(service_to_start, &DONE);
    return (TRUE);
}
#endif


static int start_DAEMON_or_SERVICE(active_h * service_to_start)
{
    /* mark it WAITING_FOR_START_DEP and wait */
    if (!mark_service(service_to_start, &WAITING_FOR_START_DEP))
    {
        W_("mark_service WAITING_FOR_START_DEP failed for service %s\n",
           service_to_start->name);
        return (FALSE);
    }
    return (TRUE);
}



static int stop_SERVICE_or_DAEMON(active_h * service_to_stop)
{
    active_h *current, *safe = NULL;

    /*
     * Okay, service is in state for stopping!
     */

    /* set stopping */
    if (!mark_service(service_to_stop, &STOP_MARKED))
        return (FALSE);

    /* also stop all service depending on service_to_stop */
    while_active_db_safe(current, safe)
    {
        if (current == service_to_stop)
            continue;

        /* dont autostop runlevels or virtuals */
        if (current->type == &TYPE_VIRTUAL)
            continue;

        /* dont stop a stopping service */
        if (current->a_state == &STOPPING)
            continue;

        /* if current depends on the one we are stopping */
        if (active_db_dep_on(current, service_to_stop) == TRUE)
            stop_service(current);
    }

    return (TRUE);
}
