
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifndef STRUCT_DATA_H
#define STRUCT_DATA_H

#include "initng_service_data_types.h"
#include "initng_list.h"

typedef struct ss_data s_data;

/*
 * This is the data struct == an initng variable, set to an service.
 */
struct ss_data
{
    /* the type of the data */
    s_entry *type;

    /* if it is a dynamic variable name, put name here */
    char *vn;

    /* and the data that it contains */
    union
    {
        /* by a char string */
        char *s;
        /* or a value */
        int i;
    } t;

    struct list_head list;
};

#define d_is(t,d) d_is_var(t,NULL,d)
int d_is_var(s_entry * type, const char *vn, struct list_head *d);

#define d_get_int(t,d) d_get_int_var(t,NULL,d)
int d_get_int_var(s_entry * type, const char *vn, struct list_head *d);

#define d_get_string(t,d) d_get_string_var(t,NULL,d)
const char *d_get_string_var(s_entry * type, const char *vn,
                             struct list_head *d);

#define d_get_next_string(t,d,l) d_get_next_string_var(t,NULL,d,l)
const char *d_get_next_string_var(s_entry * type, const char *vn,
                                  struct list_head *d, const char *last);

#define d_set(t,d) d_set_var(t, NULL, d)
void d_set_var(s_entry * type, char *vn, struct list_head *d);

#define d_set_int(t,d,v) d_set_int_var(t,NULL,d,v)
void d_set_int_var(s_entry * type, char *vn, struct list_head *d, int value);

#define d_set_string(t,d,s) d_set_string_var(t,NULL,d,s)
void d_set_string_var(s_entry * type, char *vn, struct list_head *d,
                      char *string);

#define d_set_another_string(t,d,s) d_set_another_string_var(t,NULL,d,s)
void d_set_another_string_var(s_entry * type, char *vn, struct list_head *d,
                              char *string);

int d_count_type(s_entry * type, struct list_head *d);

#define d_remove(t,d) d_remove_var(t,NULL,d)
void d_remove_var(s_entry * type, const char *vn, struct list_head *d);

void d_remove_all(struct list_head *d);

void d_copy_all(struct list_head *from, struct list_head *to);

#define while_data(current, d) list_for_each_entry_prev(current, d, list)
#define while_data_safe(current, safe, d) list_for_each_entry_prev_safe(current, safe, d, list)

#endif
