
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <termios.h>
#include <stdio.h>
#include <sys/klog.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fnmatch.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "initng_global.h"
#include "main.h"
#include "initng_signal.h"
#include "initng_handler.h"
#include "initng_execute.h"                 /* new_environ() */
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_service_data_types.h"
#include "initng_fd.h"


#define TIMEOUT 60000

static void setup_console(void)
{
    int fd;                     /* /dev/console */
    struct termios tty;

    D_("MAIN_SET_I_AM_INIT_STUFF\n");

    /* change dir to / */
    while (chdir("/") < 0)
    {
        F_("Cant chdir to \"/\"\n");
        su_login();
    }

    /* set console loglevel */
    klogctl(8, NULL, 1);


    /* enable generation of core files */
    {
        struct rlimit c = { 1000000, 1000000 };
        setrlimit(RLIMIT_CORE, &c);
    }

    reboot(RB_DISABLE_CAD);                 /* Disable Ctrl + Alt + Delete */

    /* open the console */
    if (g.dev_console)
    {
        fd = open(g.dev_console, O_RDWR | O_NOCTTY);
    }
    else
    {
        fd = open(INITNG_CONSOLE, O_RDWR | O_NOCTTY);
    }

    /* Try to open the console, but don't control it */
    if (fd > 0)
    {
        D_("Opened " INITNG_CONSOLE ". Setting terminal options.\n");
        ioctl(fd, KDSIGACCEPT, SIGWINCH);   /* Accept signals from 'kbd' */
        close(fd);                          /* Like Ctrl + Alt + Delete signal? */
    }
    else
    {
        D_("Failed to open " INITNG_CONSOLE ". Setting options anyway.\n");
        ioctl(0, KDSIGACCEPT, SIGWINCH);    /* Accept signals from 'kbd' */
    }

    /* TODO: this block may be incorrect or incomplete */
    /* Q: What does this block really do? */

    /*
     * TODO: /dev/console may still be open from before. Also, if it
     * fails to open, why try to finish the block?
     */
    if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) < 0)
    {
        F_("main(): can't open /dev/console.\n");
    }

    (void) tcgetattr(fd, &tty);

    tty.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
    tty.c_cflag |= HUPCL | CLOCAL | CREAD;

    tty.c_cc[VINTR] = 3;                    /* ctrl('c') */
    tty.c_cc[VQUIT] = 28;                   /* ctrl('\\') */
    tty.c_cc[VERASE] = 127;
    tty.c_cc[VKILL] = 24;                   /* ctrl('x') */
    tty.c_cc[VEOF] = 4;                     /* ctrl('d') */
    tty.c_cc[VTIME] = 0;
    tty.c_cc[VMIN] = 1;
    tty.c_cc[VSTART] = 17;                  /* ctrl('q') */
    tty.c_cc[VSTOP] = 19;                   /* ctrl('s') */
    tty.c_cc[VSUSP] = 26;                   /* ctrl('z') */

    /*
     *      Set pre and post processing
     */
    tty.c_iflag = IGNPAR | ICRNL | IXON | IXANY;
    tty.c_oflag = OPOST | ONLCR;
    tty.c_lflag = ISIG | ICANON | ECHO | ECHOCTL | ECHOPRT | ECHOKE;

    /*
     *      Now set the terminal line.
     *      We don't care about non-transmitted output data
     *      and non-read input data.
     */
    (void) tcsetattr(fd, TCSANOW, &tty);
    (void) tcflush(fd, TCIOFLUSH);
    (void) close(fd);

}


/*
 * %%%%%%%%%%%%%%%%%%%%   main ()   %%%%%%%%%%%%%%%%%%%%
 */
#ifdef BUSYBOX
int init_main(int argc, char *argv[], char *env[])
{
#else
int main(int argc, char *argv[], char *env[])
{
#endif
    time_t last;                /* save the time here */

#ifdef DEBUG
    int loop_counter = 0;       /* a common used for counter */
#endif

    S_;

    /* get the time */
    last = time(NULL);

    /* initialise global variables */
    initng_new(argc, argv, env);

    /* set i_am_init if pid == 1 */
    if (getpid() == 1)
        g.i_am_init = TRUE;

    /* parse all options, set in argv */
    initng_parse_argv(argv);

    /* if this is real init */
    if (g.i_am_init)
    {
        /* when last service stoped, offer a sulogin */
        g.when_out = THEN_SULOGIN;
        if (!g.runlevel)
            initng_set_runlevel("default");

        if (!g.hot_reload)
        {
            /* static function abowe, initziates the system */
            setup_console();
        }

    }
    else
    {
        /* when last service stoped, quit initng */
        g.when_out = THEN_QUIT;
        initng_set_runlevel("fake-default");
    }

    D_("MAIN_LOAD_MODULES\n");
    /* Load modules, if fails - launch sulogin and then try again */
    if (!initng_load_all_modules())
    {
        /* if we can't load modules, revert to single-user login */
        su_login();
    }

    /* set title of initng, can be watched in ps -ax */
    set_proc_title("initng [%s]", g.runlevel);

    /* Configure signal handlers */
    (void) enable_signals();

    /* change system state */
    set_sys_state(STATE_STARTING);

    /* make sure this is not a hot reload */
    if (!g.hot_reload)
    {
        /* first start all services, prompted at boot with +daemon/test */
        start_extra_services();

        /* try load the default service, if it fails - launch sulogin and try again */
        if (!start_new_service_named(g.runlevel))
        {
            F_("Failed to load runlevel (%s)!\n", g.runlevel);
            su_login();
        }
    }

    D_("MAIN_GOING_MAIN_LOOP\n");
    /* %%%%%%%%%%%%%%%   MAIN LOOP   %%%%%%%%%%%%%%% */
    for (;;)
    {
        D_("MAIN_LOOP: %i\n", loop_counter++);

        /* Update current time, save this in global so we dont need to call time() that often. */
        g.now = time(NULL);

        /* if this isnt a normal timeout loop */
        if (g.sleep_seconds != TIMEOUT)
        {
            initng_set_interrupt();
            /* Reset the sleep counter, to sleep forever. */
            g.sleep_seconds = TIMEOUT;
        }


        /*
         * If it has gone more then 1h sence last
         * main loop, or if clock has gon backwards,
         * reset it.
         */
        if ((g.now - last) >= 3600 || (g.now - last) < 0)
        {
            D_(" Clock skew, time have changed over one hour, in one mainloop, compensating 3600 seconds.\n");
            active_db_compensate_time(g.now - last);
            initng_compensate_time_call(g.now - last);
            last += (g.now - last);
        }
        /* put last time, to current time */
        last = g.now;



        /* Check for sigchld */
        if (initng_signal_got_sigchld)
        {
            D_("There are dead childs to handle!\n");
            initng_signal_got_sigchld = FALSE;
            handle_sigchild();
        }

        /* Check for sig alarm */
        if (initng_signal_got_alarm)
        {
            D_("Got alarm set, checking timers plugin ...\n");
            initng_signal_got_alarm = FALSE;
            /* Check for timed out stopping services for SIGKILL */
            term_stopping_services();
            /* Check timers plugins */
            initng_plugin_alarm();
        }

        /* If we got an interrupt load ctrlaltdel */
        if (initng_signal_got_ctrl_alt_del)
        {
            D_("I got ctrl+alt+del!\n");
            initng_signal_got_ctrl_alt_del = FALSE;

            /* what to do when there is no services left */
            if (g.i_am_init)
                g.when_out = THEN_REBOOT;
            else
                g.when_out = THEN_QUIT;

            /* stop all services */
            stop_all();
        }

        /* If we got a keyboard request load kbreq */
        if (initng_signal_got_keyboard_request)
        {
            D_("I got a keyboard request!\n");
            initng_signal_got_keyboard_request = FALSE;
            if (!start_new_service_named("kbreq"))
                F_("keyboard requires service \"kbreq\", could not be launched.\n");
        }


        /* 
         * When g.interrupt is set, we are going to launch all handlers,
         * for all services, its in here there is acutally happen
         * something.
         *
         * It is a while loop, becouse it will probably reset g.interrupt
         * many times, so this would be like a mini main-loop.
         */
        while (g.interrupt == TRUE)
        {
            D_("Resetting g.interrupt = FALSE\n");
            g.interrupt = FALSE;
            initng_run_interrupt_handlers();
            D_("mini loop (will continue as long as interrupt is set)\n");
        }

        /* If there is modules to unload, handle this */
        if (g.modules_to_unload == TRUE)
        {
            g.modules_to_unload = FALSE;
            D_("There is modules to unload!\n");
            initng_unload_marked_modules();
        }


        /* if interupt is set, dont waits cpu on more checks */
        if (g.interrupt == TRUE)
            continue;

        /*
         * After here i put things that should be done every main
         * loop, g.interrupt or not ...
         */

        /*
         * call the main_hooks in all plugins.
         * OBS! This part in mainloop is expensive.
         */
        {
            s_call *current, *q = NULL;

            while_call_db_safe(current, q)
            {
                if (current->ptype != MAIN)
                    continue;
                (*current->c.main) ();
            }
        }


        /*
         * Check if there are any running processes left, otherwise quit, when_out try to do something when this happends.
         *
         * This check is also expensive.
         */
        if (ready_to_quit() == TRUE)
        {
            set_sys_state(STATE_ASE);
            initng_set_interrupt();
            P_(" *** Last service has quit. ***\n");
            when_out();
            continue;
        }

        /* if interupt is set, dont waits cpu on a poll */
        if (g.interrupt == TRUE)
            continue;

        /* sleep the seconds set in g.sleep_seconds */
        initng_plugin_fd_poll(g.sleep_seconds);
        D_("g.sleep_seconds: %i elapsed!\n", g.sleep_seconds);
    }                                       /* End main loop */
}
